library(orthoDr)
library(MASS)
library(ManifoldOptim)
library(RcppArmadillo)
library(pracma)


### orthoDr Surv

# Generate Some Survival Data
set.seed(1)
N = 350; P = 6; dataX = matrix(rnorm(N*P), N, P)
failEDR = as.matrix(cbind(c(1, 1, 0, 0, 0, 0, rep(0, P-6)),
                          c(0, 0, 1, -1, 0, 0, rep(0, P-6))))
censorEDR = as.matrix(c(0, 1, 0, 1, 1, 1, rep(0, P-6)))
T = exp(-2.5 + dataX %*% failEDR[,1] + 0.5*(dataX %*%
                                                 +  failEDR[,1])*(dataX %*% failEDR[,2]) + 0.25*log(-log(1-runif(N))))
C = exp( -0.5 + dataX %*% censorEDR  + log(-log(1-runif(N))))
Y = pmin(T, C)
Censor = (T < C)

# fit the model
orthoDr.fit = orthoDr_surv(dataX, Y, Censor, ndr = 2)
orthoDr.fit

# Calculate the distance to the true parameters
distance(failEDR, orthoDr.fit$B, "dist")

# Compare with the counting process inverse regression model
orthoDr.fit1 = orthoDr_surv(dataX, Y, Censor, method = "dn", ndr = 2)
distance(failEDR, orthoDr.fit1$B, "dist")


### orthoDr for regression model: semiparametric phd (Ma and Zhu, 2012)

set.seed(1)
N = 100; P = 4; dataX = matrix(rnorm(N*P), N, P)
Y = -1 + dataX[,1] + rnorm(N)
orthoDr_reg(dataX, Y, ndr = 1, method = "phd")


### Paralleled gradient approximation through OpenMP

Table1 = matrix(NA,2,2) # Computational Cost of Different Numbers of Cores
rownames(Table1) = c("ncore = 1", "ncore = 4")
colnames(Table1) = c("P = 6", "P = 12")

nsim = 10 # number of simulations
N = 350
ndr = 2

for(i in 1:2){ # loop for ncores

	set.seed(1) # replicate data for different ncores
	
	for(j in 1:2){ # loop for P
	
		P = c(6,12)[j]
		temp = rep(NA, nsim)
		
		for(k in 1:nsim){
			dataX = matrix(rnorm(N*P), N, P)

			failEDR = as.matrix(cbind(c(1, 1, 0, 0, 0, 0, rep(0, P-6)),
									  c(0, 0, 1, -1, 0, 0, rep(0, P-6))))

			censorEDR = as.matrix(c(0, 1, 0, 1, 1, 1, rep(0, P-6)))

			T = exp(-2.5 + dataX %*% failEDR[,1] + 0.5*(dataX %*%
				  +  failEDR[,1])*(dataX %*% failEDR[,2]) + 0.25*log(-log(1-runif(N))))
			C = exp( -0.5 + dataX %*% censorEDR  + log(-log(1-runif(N))))
			Y = pmin(T, C)
			Censor = (T < C)
			
			t0 = Sys.time()
			dn.fit = orthoDr_surv(dataX, Y, Censor, method = "dn", ndr = ndr, ncore = c(1,4)[i], maxitr = 1000, control = list(ftol = 1e-6))
			t1 = Sys.time()
			
			temp[k] = difftime(t1, t0, units = "secs")
		}
		Table1[i, j] = mean(temp)
	}
}

Table1


### General Solver Brockett Problem

n = 150; p = 5; set.seed(1)

X <- matrix(rnorm(n*n), nrow=n)
X <- X + t(X)
D <- diag(p:1, p)

f1 <- function(B, X, D) { Trace( t(B) %*% X %*% B %*% D ) }
g1 <- function(B, X, D) { 2 * X %*% B %*% D }
b1 = gramSchmidt(matrix(rnorm(n*p), nrow=n, ncol=p))$Q

res2 = ortho_optim(b1, fn = f1, grad = g1, X, D)
head(res2$B)
res2$fn

# Generate Tables for comparison with the ManifoldOptim package


set.seed(1)
method = c('LRBFGS', 'LRTRSR1', 'RBFGS','RTRSR1')  # 'RBFGS','RTRSR1' these two method runs significantly slower. If needed, please add them to the method vector

nsim = 100 # number of simulations
n = 150
p = 5
maxitr = 750

Table1 = matrix(NA, nsim, length(method)+1) # Table of Running Time
colnames(Table1) = c(method, "orthoDr")
Table2 = list() # list of Functional Value
F0 = rep(NA, nsim)

for(k in 1:nsim)
{
  Temp_time = rep(NA, length(method)+1)
  Temp_fval = matrix(NA, length(method)+1, maxitr)
  rownames(Temp_fval) = c(method, "orthoDr")
  colnames(Temp_fval) = 1:maxitr

	X <- matrix(rnorm(n*n), nrow=n)
	X <- X + t(X)
	D <- diag(p:1, p)

	# Functions for ManifoldOptim
	tb <- function(b) { matrix(b, n, p) }
	f <- function(b) { B <- tb(b); Trace( t(B) %*% X %*% B %*% D ) }
	g <- function(b) { B <- tb(b); 2 * X %*% B %*% D }

	mod <- Module("ManifoldOptim_module", PACKAGE = "ManifoldOptim")
	prob <- new(mod$RProblem, f, g)

	# Set up Initial Value and Estimate
	b0 <- as.numeric(orthonorm(matrix(rnorm(n*p), nrow=n, ncol=p)))
	mani.params <- get.manifold.params(IsCheckParams = FALSE)
	mani.defn <- get.stiefel.defn(n, p)

	for(l in 1:length(method)){
	  
	  # trace fun value 
	  solver.params <- get.solver.params(DEBUG = 2, Tolerance = 1e-15, Max_Iteration = maxitr, IsCheckParams = FALSE, IsCheckGradHess = FALSE)	  
		res <- manifold.optim(prob, mani.defn, method = method[l],
							  mani.params = mani.params, solver.params = solver.params,
							  x0 = b0)
		
		Temp_fval[l, ] = res$funSeries[2:(maxitr+1)]
		
		# compuational time
		solver.params <- get.solver.params(DEBUG = 0, Tolerance = 1e-15, Max_Iteration = maxitr, IsCheckParams = FALSE, IsCheckGradHess = FALSE)	  
		time.start <- Sys.time()
		res <- manifold.optim(prob, mani.defn, method = method[l],
		                      mani.params = mani.params, solver.params = solver.params,
		                      x0 = b0)
		time.end <- Sys.time()
		
		Temp_time[l] = difftime(time.end, time.start, units = "secs")
	}
		
	# Functions for orthoDr
	f1 <- function(B, X, D) { Trace( t(B) %*% X %*% B %*% D ) }
	g1 <- function(B, X, D) { 2 * X %*% B %*% D }
	
	b1 = gramSchmidt(matrix(b0, n, p))$Q
	
	grad_0 = sqrt(sum(abs(g1(b1, X, D))^2)) # ManifoldOptim uses G_1/G_0, but ours use G_1. We adjust the criteria to match them
	
	time.start <- Sys.time()
	res2 <- ortho_optim(b1, fn = f1, grad = g1, X, D, maxitr = maxitr, control = list(gtol = 1e-15, ftol = 1e-15, btol = 1e-15))
	time.end <- Sys.time()
	Temp_time[length(method)+1] = difftime(time.end, time.start, units = "secs")
	Temp_fval[length(method)+1, ] = res2$fn_Seq
  
  Table1[k, ] = Temp_time
  Table2[[k]] = Temp_fval
  F0[k] = f1(b1, X, D)
  
  cat(paste("-------------------  run", k, "-------------------------\n"))
}


PlotPath = list()
for (k in 1:length(Table2))
{
  Temp = Table2[[k]]
  Temp[Temp == 0] = NA
  PlotPath[[k]] = log((Temp - min(Temp, na.rm = TRUE)))
}

PlotPath = t(Reduce("+", PlotPath) / length(PlotPath))

use = c(1:length(method), 5)

par(mar = c(4.5, 5, 1, 1))
matplot(PlotPath, type = "l", lty = use, col = use, lwd = 2, yaxt = "n",
        xlab = "iteration", ylab = "functional value - minimum", cex.lab = 2)
legend("topright", c(method, "orthoDr"), lty = use, col = use, lwd =2, cex = 1.5)
axis(2, at = log(10^{c(-7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3)}), labels = c("1e-7", "1e-6", "1e-5", "1e-4", "1e-3", "1e-2", "1e-1", "1", "1e1", "1e2", "1e3"))

round(colMeans(Table1, na.rm = TRUE), 3)


### Case Example
library(rgl)

# Loading Data Downloaded
concrete_data = read.csv("Concrete_Data.csv")
X = as.matrix(concrete_data[,1:8])
Y = as.matrix(concrete_data[,9])
colnames(X) = c("Cement", "Blast Furnace Slag", "Fly Ash", "Water", "Superplasticizer", "Coarse Aggregate", "Fine Aggregate", "Age")
# Fit the model
result = orthoDr_reg(X, Y, ndr = 2, method = "sir", maxitr = 1000, keep.data = TRUE, verbose = TRUE)
rownames(result$B) = colnames(X)
result$B

# Plot
BX = X %*% result$B
BX = scale(BX)/result$bw/sqrt(2)
plot3d(BX[,1], BX[, 2], Y, col="red", size=5, xlab='',ylab='',zlab='')

gridX = seq(min(BX[,1]), max(BX[,1]), length = 100)
gridY = seq(min(BX[,2]), max(BX[,2]), length = 100)


# Get the surfacd from the reduce directions
predict_regDir <- function(object, testBX, ...)
{
  # transform the covariates into the same scale
  X = object$x

  BX = X %*% object$B
  BX = scale(BX)/object$bw/sqrt(2)

  testKernel = orthoDr:::KernelDist_cross(testBX, BX)

  pred = apply(testKernel, 1, function(w, x) weighted.mean(x, w), object$y)

  return(list("pred" = pred))
}

gridYX = expand.grid(gridY,gridX)
newBX = cbind(gridYX[,2], gridYX[,1])
S = matrix(predict_regDir(result, newBX)$pred, 100, 100, byrow = TRUE)
surface3d(gridX, gridY, S, col = 'grey', alpha = 0.4, add = T)






